--[[
	Copyright (C) 2007 Nymbia
	Copyright (C) 2007 Chris "kergoth" Larson <clarson@kergoth.com> - scale bits

	This program is free software; you can redistribute it and/or modify
	it under the terms of the GNU General Public License as published by
	the Free Software Foundation; either version 2 of the License, or
	(at your option) any later version.

	This program is distributed in the hope that it will be useful,
	but WITHOUT ANY WARRANTY; without even the implied warranty of
	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
	GNU General Public License for more details.

	You should have received a copy of the GNU General Public License along
	with this program; if not, write to the Free Software Foundation, Inc.,
	51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
]]

PT3Bar = AceLibrary("AceAddon-2.0"):new("AceEvent-2.0", "AceDB-2.0", "AceConsole-2.0")
PT3Bar:RegisterDB("PT3BarDB")
local PT = AceLibrary("PeriodicTable-3.0")
local _G = getfenv(0)

local scale, uiscale
local popupparent
local buttons
local popupbuttons = {}
local size = 36
local binds
local bindparent = CreateFrame("Frame",nil,UIParent)
function PT3Bar:InputSets(t,b)
	buttons = t
	binds = b
	self.InputSets = nil
end
local function positionbutton(button, k)
	local db = PT3Bar.db.profile[k]
	button:ClearAllPoints()
	button:SetPoint("BOTTOMLEFT", UIParent, "BOTTOMLEFT", db.x and db.x/scale or (200+(k*(size+4))), (db.y or 200)/scale)
end
local settables = {}
local locked = true
local options = {
	type = 'group',
	args = {
		lock = {
			type = 'toggle',
			name = "Lock",
			desc = "Lock",
			get = function()
				return locked
			end,
			set = function(v)
				locked = v
				if v then
					for _,v in pairs(buttons) do
						v.moveoverlay:Hide()
					end
				else
					for _,v in pairs(buttons) do
						positionbutton(v, tonumber(v:GetName():match("PT3Bar(%d+)")))
						v:Show()
						v.moveoverlay:Show()
					end
				end
				PT3Bar:UpdateAllButtons()
			end,
		},
		scale = {
			type = 'range', min = 0.1, max = 2.0, step = 0.1,
			name = "Scale",
			desc = "Button Scale",
			get = function()
				return PT3Bar.db.profile.scale
			end,
			set = function(v)
				scale = uiscale * v
				PT3Bar.db.profile.scale = v
				for _,b in pairs(buttons) do
					local k = tonumber(b:GetName():match("PT3Bar(%d+)"))
					b:SetScale(v)
					positionbutton(b, k)
					
				end
				
				PT3Bar:Compress()
			end,
		},
		snap = {
			type = 'group',
			name = "Snap",
			desc = 'Snap',
			args = {
				h = {
					type = 'execute',
					name = "Horizontal",
					desc = "Snaps the buttons Horizontally",
					func = function()
						local x1, y1
						for _,frame in pairs(buttons) do
							local num = tonumber(frame:GetName():match("PT3Bar(%d+)"))
							if num == 1 then
								x1 = frame:GetLeft()
								y1 = frame:GetBottom()
							end
						end
						for _,frame in pairs(buttons) do
							local num = tonumber(frame:GetName():match("PT3Bar(%d+)"))
							if num ~= 1 then
								frame:ClearAllPoints()
								frame:SetPoint("BOTTOMLEFT", UIParent, "BOTTOMLEFT", x1+((size+4)*(num-1)), y1)
								PT3Bar.db.profile[num].x = frame:GetLeft() * scale
								PT3Bar.db.profile[num].y = frame:GetBottom() * scale
							end
						end
						
						PT3Bar:Compress()
					end
				},
				v = {
					type = 'execute',
					name = "Vertical",
					desc = "Snaps the buttons Vertically",
					func = function()
						local x1, y1
						for _,frame in pairs(buttons) do
							local num = tonumber(frame:GetName():match("PT3Bar(%d+)"))
							if num == 1 then
								x1 = frame:GetLeft()
								y1 = frame:GetBottom()
							end
						end
						for _,frame in pairs(buttons) do
							local num = tonumber(frame:GetName():match("PT3Bar(%d+)"))
							if num ~= 1 then
								frame:ClearAllPoints()
								frame:SetPoint("BOTTOMLEFT", UIParent, "BOTTOMLEFT", x1, y1-((size+4)*(num-1)))
								PT3Bar.db.profile[num].x = frame:GetLeft() * scale
								PT3Bar.db.profile[num].y = frame:GetBottom() * scale
							end
						end
						
						PT3Bar:Compress()
					end
				}
			},
		},
		compress = {
			type = 'group',
			name = "Compress",
			desc = 'Horizontal/vertical compressing towards the first button when the bar is locked',
			args = {
				compressH = {
					type = 'toggle',
					name = "Horizontal",
					desc = "Compresses the buttons horizontally",
					get = function()
						return PT3Bar.db.profile.compressH
					end,
					set = function(v)
						PT3Bar.db.profile.compressH = v
						
						if v then
							PT3Bar.db.profile.compressV = false
							PT3Bar:UpdateAllButtons()
						else
							for _,b in pairs(buttons) do
								positionbutton(b, tonumber(b:GetName():match("PT3Bar(%d+)")))
							end							
						end
					end,					
				},
				compressV = {
					type = 'toggle',
					name = "Vertical",
					desc = "Compresses the buttons vertically",
					get = function()
						return PT3Bar.db.profile.compressV
					end,
					set = function(v)
						PT3Bar.db.profile.compressV = v
						
						if v then
							PT3Bar.db.profile.compressH = false
							PT3Bar:UpdateAllButtons()
						else
							for _,b in pairs(buttons) do
								positionbutton(b, tonumber(b:GetName():match("PT3Bar(%d+)")))
							end							
						end
					end,					

				},
			},
		},		
		tooltip = {
			type = 'toggle',
			name = "Tooltip",
			desc = "Tooltip",
			get = function()
				return PT3Bar.db.profile.tooltip
			end,
			set = function(v)
				PT3Bar.db.profile.tooltip = v
			end,
		},
		showbest = {
			type = 'text',
			name = "ShowBest",
			desc = "Show best item in button",
			get = function()
				if PT3Bar.db.profile.showbest then
					return "Best to Worst"
				else
					return "Worst to Best"
				end
			end,
			set = function(v)
				if v == "Best to Worst" then
					PT3Bar.db.profile.showbest = true
				else
					PT3Bar.db.profile.showbest = false
				end
				if PT3Bar1 then
					PT3Bar:UpdateAllButtons()
				end
			end,
			validate = { "Best to Worst", "Worst to Best" },
		},
		popupdirection = {
			type = 'text',
			name = "Popup Direction",
			desc = "Determine the direction of the popup buttons",
			get = function()
				local v = PT3Bar.db.profile.growdir
				if v == 'u' then
					return "Up"
				elseif v == 'd' then
					return "Down"
				elseif v == "l" then
					return "Left"
				else
					return "Right"
				end
			end,
			set = function(v)
				if v == "Up" then
					PT3Bar.db.profile.growdir = 'u'
				elseif v == "Down" then
					PT3Bar.db.profile.growdir = 'd'
				elseif v == "Left" then
					PT3Bar.db.profile.growdir = 'l'
				else
					PT3Bar.db.profile.growdir = 'r'
				end
				PT3Bar:ApplyPopupGrowDirection()
			end,
			validate = {"Up","Down","Left","Right"},
		},
	}
}
function PT3Bar:OnInitialize()
	self:RegisterChatCommand({"/pt3bar"}, options)
	self:RegisterDefaults('profile', {
		tooltip = true,
		showbest = true,
		compressH = false,
		compressV = false,
		growdir = "u",
		scale = 1,
		['*'] = {}
	})

	if tonumber(GetCVar("useUiScale")) == 1 then
		uiscale = tonumber(GetCVar("uiScale"))
	else
		uiscale = 1
	end
end

function PT3Bar:OnEnable(first)
	scale = self.db.profile.scale * uiscale
	self:RegisterEvent("CVAR_UPDATE", function(cvarname, value)
		if cvarname == "uiScale" then
			uiscale = tonumber(value)
		elseif cvarname == "useUiScale" then
			if tonumber(value) == 1 then
				uiscale = tonumber(GetCVar("uiScale"))
			else
				uiscale = 1
			end
		end
		scale = self.db.profile.scale * uiscale
	end)

	if first then
		popupparent = CreateFrame("Frame", "PT3BarPopupParent", UIParent, "SecureStateHeaderTemplate")
		local newbuttons = {}
		for k,v in ipairs(buttons) do
			local button = CreateFrame("CheckButton", "PT3Bar"..k, UIParent, "ActionBarButtonTemplate, SecureAnchorEnterTemplate")
			
			button:SetMovable(true)
			button:SetClampedToScreen(true)
			
			button:SetScript("OnReceiveDrag", nil)
			button:SetScript("OnDragStart", nil)
			button:UnregisterAllEvents()
			button:SetAttribute("type", "item")
			button:SetAttribute("*unit2", "player")
			button.icon = _G["PT3Bar"..k.."Icon"]
			button.icon:SetVertexColor(1,1,1,1)
			button.count = _G["PT3Bar"..k.."Count"]
			button.cooldown = _G["PT3Bar"..k.."Cooldown"]
			_G["PT3Bar"..k.."Name"]:Hide()
			_G["PT3Bar"..k.."HotKey"]:Hide()

			button:SetWidth(size)
			button:SetHeight(size)
			_G["PT3Bar"..k.."NormalTexture"]:SetWidth((size/36)*66)
			_G["PT3Bar"..k.."NormalTexture"]:SetHeight((size/36)*66)
			button:SetScale(self.db.profile.scale)
			positionbutton(button, k)
			
			local moveoverlay = CreateFrame("Frame",nil,button)
			button.moveoverlay = moveoverlay

			moveoverlay:SetScript("OnMouseDown", function(frame, mousebtn)
				if not locked then
					if mousebtn == 'RightButton' then
						local v = not button.hidden
						button.hidden = v
						self.db.profile[k].hidden = v
						self:SetButtonHidden(button, v)
						
						moveoverlay:SetScript("OnMouseUp", nil)
					else
						button:StartMoving()
						moveoverlay:SetScript("OnMouseUp", function()
							button:StopMovingOrSizing()
							self.db.profile[k].x = button:GetLeft() * scale
							self.db.profile[k].y = button:GetBottom() * scale
							PT3Bar:Compress()
						end)
					end
				end
			end)

			moveoverlay:EnableMouse(true)
			moveoverlay:SetFrameStrata("HIGH")
			moveoverlay:SetWidth(size)
			moveoverlay:SetHeight(size)
			moveoverlay:SetPoint("TOPLEFT",button,"TOPLEFT",0,0)
			moveoverlay:Hide()

			button.hidden = self.db.profile[k].hidden
			self:SetButtonHidden(button, button.hidden)
			
			newbuttons[v] = button
			settables[v] = PT(v)
			
			
			popupbuttons[k] = {}
			--local popupparent = CreateFrame("Frame", "PT3BarPopupParent"..k, button, "SecureStateHeaderTemplate")
			
			--button.popupparent = popupparent
			
			
			button:SetAttribute("anchorchild", popupparent)
			button:SetAttribute("*childraise-OnEnter", true)
			button:SetAttribute("*childstate-OnEnter", "enter"..k)
			button:SetAttribute("*childstate-OnLeave", "leave"..k)
			button:SetAttribute("checkselfcast", true)
			button:SetAttribute("newstate", k..":0")
			
			popupparent:SetAttribute("statemap-anchor-enter"..k, "*:"..k)
			popupparent:SetAttribute("statemap-anchor-leave"..k, ";")
			popupparent:SetAttribute("delaystatemap-anchor-leave"..k, k..":0")
			popupparent:SetAttribute("delaytimemap-anchor-leave"..k, k..":0.5")
			popupparent:SetAttribute("delayhovermap-anchor-leave"..k, k..":true")
					
			local popupbutton1 = self:GetPopupButton(k, 1)
			popupbutton1:SetScript("OnShow", function()
				self:SetTooltip(button)
			end)
			popupbutton1:SetScript("OnHide", function()
				GameTooltip:Hide()
			end)
		end
		buttons = newbuttons

		popupparent:SetPoint("TOPLEFT", button, "TOPLEFT")
		
		popupparent:SetAttribute("state", "0")
		popupparent:SetAttribute("hidestates", "0")

	end
	ClearOverrideBindings(bindparent)
	for k,v in pairs(binds) do
		if v then
			SetOverrideBindingClick(bindparent, false, v, "PT3Bar"..k)
		end
	end
	for _,v in pairs(buttons) do
	        v:Show()
	end
	self:UpdateAllButtons()
	self:ScheduleRepeatingEvent(self.UpdateCooldowns, 2, self)
	self:RegisterBucketEvent("BAG_UPDATE", 2)
	
end

function PT3Bar:GetPopupButton(buttonnum, popupnum)
	local b = popupbuttons[buttonnum][popupnum]
	if b then
		if not b:GetParent() then
			popupparent:SetAttribute("addchild", b)
		end
		return b
	end
	local popup = CreateFrame("CheckButton", "PT3Bar"..buttonnum.."Popup"..popupnum, UIParent, "ActionBarButtonTemplate")
	--clear some annoying inherited stuff
	popup:SetScript("OnReceiveDrag", nil)
	popup:SetScript("OnDragStart", nil)
	popup:UnregisterAllEvents()
	popup:SetAttribute("type", "item")
	popup:SetAttribute("*unit2", "player")
	popup:SetAttribute("newstate", buttonnum..":0")
	popup.icon = _G["PT3Bar"..buttonnum.."Popup"..popupnum.."Icon"]
	popup.icon:SetVertexColor(1,1,1,1)
	popup.count = _G["PT3Bar"..buttonnum.."Popup"..popupnum.."Count"]
	popup.cooldown = _G["PT3Bar"..buttonnum.."Popup"..popupnum.."Cooldown"]
	_G["PT3Bar"..buttonnum.."Popup"..popupnum.."Name"]:Hide()
	_G["PT3Bar"..buttonnum.."Popup"..popupnum.."HotKey"]:Hide()
	popup:SetWidth(size)
	popup:SetHeight(size)
	_G["PT3Bar"..buttonnum.."Popup"..popupnum.."NormalTexture"]:SetWidth((size/36)*66)
	_G["PT3Bar"..buttonnum.."Popup"..popupnum.."NormalTexture"]:SetHeight((size/36)*66)
	--popup:SetScale(self.db.profile.scale)
	popup:SetScale(0.9)
	
	popupparent:SetAttribute("addchild", popup)
	popup:SetFrameStrata('HIGH')
	
	popup:SetScript("OnEnter", function()
		self:SetTooltip(popup)
	end)
	popup:SetScript("OnLeave", function()
		GameTooltip:Hide()
	end)
	
	if cyCircled and cyCircled_PT3Bar then
		cyCircled_PT3Bar:ApplySkin()
	end
	
	popup:Hide()
	popup:SetAttribute("showstates", tostring(buttonnum))
	
	popupbuttons[buttonnum][popupnum] = popup
	self:ApplyPopupGrowDirection()
	return popup
end

function PT3Bar:OnProfileEnable()
	for _, frame in pairs(buttons) do
		frame:ClearAllPoints()
		local k = tonumber(frame:GetName():match("PT3Bar(%d+)"))
		positionbutton(frame, k)
		self:ApplyPopupGrowDirection()
	end
end

function PT3Bar:ApplyPopupGrowDirection()
	local point, relpoint, v, n
	if self.db.profile.growdir == "u" then
		point = "BOTTOM"
		relpoint = "TOP"
		v = true
	elseif self.db.profile.growdir == "d" then
		point = "TOP"
		relpoint = "BOTTOM"
		v = true
		n = true
	elseif self.db.profile.growdir == "l" then
		point = "RIGHT"
		relpoint = "LEFT"
		n = true
	else --"r"
		point = "LEFT"
		relpoint = "RIGHT"
	end
	for buttonnum, t in ipairs(popupbuttons) do
		for popupnum, button in ipairs(t) do
			button:ClearAllPoints()
			button:SetPoint(point, _G["PT3Bar"..buttonnum], relpoint, not v and size*(popupnum-1) * (n and -1 or 1) or 0, v and size*(popupnum-1) * (n and -1 or 1)  or 0)
		end
	end
end

function PT3Bar:OnDisable()
	ClearOverrideBindings(bindparent)
	for _,v in pairs(buttons) do
		v:Hide()
	end
end

function PT3Bar:BAG_UPDATE()
	if InCombatLockdown() then
		self:UpdateCountsAndAlphas()
		self:CancelAllCombatSchedules()
		self:ScheduleLeaveCombatAction(self.UpdateAllButtons, self)
	else
		self:UpdateAllButtons()
	end
end

function PT3Bar:UpdateCountsAndAlphas()
	for set,button in pairs(buttons) do
		if not button.item then
			return
		end
		local count = GetItemCount(button.item)
		if count == 0 then
			button:SetAlpha(0.2)
		end
		button.count:SetText(count)
	end
	for _, t in ipairs(popupbuttons) do
		for _, button in ipairs(t) do
			if not button.item then
				return
			end
			local count = GetItemCount(button.item)
			if count == 0 then
				button:SetAlpha(0.2)
			end
			button.count:SetText(count)
		end
	end
end

local emptyButtons = {}
local addPopups = {}

function PT3Bar:UpdateAllButtons()

	for k in pairs( emptyButtons ) do
		emptyButtons[k] = nil
	end

	for set,button in pairs(buttons) do
		local _, firstv = next(settables[set])
		local currvalue = self.db.profile.showbest and 0 or 100000
		local curritem
		local currcount
		local popupincrement = 1
		local buttonnum = tonumber(button:GetName():match("PT3Bar(%d+)$"))
		local state = popupparent:GetAttribute('state')
		if type(firstv) == "table" then
			--multiset
			for _,t in ipairs(settables[set]) do
				for k,v in pairs(t) do
					v = tonumber(v)
					k = tonumber(k)
					local count = k and GetItemCount(k) or 0
					if count > 0 then
						if v and ((self.db.profile.showbest and v > currvalue) or ((not self.db.profile.showbest) and v < currvalue)) then
							curritem = k
							currcount = count
							currvalue = v
						end
						--self:UpdateButton(self:GetPopupButton(buttonnum, popupincrement),k,count)
						--popupincrement = popupincrement + 1
						addPopups[k] = count
					end
				end
			end
		else
			--normal set
			for k,v in pairs(settables[set]) do
				v = tonumber(v)
				k = tonumber(k)
				local count = k and GetItemCount(k) or 0
				if count > 0 then
					if v and ((self.db.profile.showbest and v > currvalue) or ((not self.db.profile.showbest) and v < currvalue)) then
						curritem = k
						currcount = count
						currvalue = v
					end
					--self:UpdateButton(self:GetPopupButton(buttonnum, popupincrement),k,count)
					--popupincrement = popupincrement + 1
					addPopups[k] = count
				end
			end
		end
		if curritem then
			addPopups[curritem] = nil
		end
		for k,v in pairs( addPopups ) do
			self:UpdateButton(self:GetPopupButton(buttonnum, popupincrement),k,v)
			popupincrement = popupincrement + 1
			addPopups[k] = nil
		end
		if curritem then
			self:UpdateButton(button,curritem,currcount)
		else
			if popupincrement > 1 then
				local p = self:GetPopupButton(buttonnum, 1)
				self:UpdateButton(button,p.item,tonumber(p.count:GetText()))
			else
				if locked then
					button:Hide()
					emptyButtons[buttonnum] = true
				end
				button.icon:SetTexture("Interface\\Icons\\INV_Misc_QuestionMark")
			end
		end
		if locked and button.hidden then
			emptyButtons[buttonnum] = true
		end
		local popuptbl = popupbuttons[buttonnum]
		while popuptbl[popupincrement] do
			local f = popuptbl[popupincrement]
			f:SetParent(nil)
			f:Hide()
			f:SetAttribute("state", "0")
			f.item = nil
			popupincrement = popupincrement + 1
		end
		popupparent:SetAttribute('state', state)
	end
	
	PT3Bar:Compress()
end

function PT3Bar:Compress()

	if PT3Bar.db.profile.compressH or PT3Bar.db.profile.compressV then
	
		local xPrev, yPrev
	
		for _,frame in pairs(buttons) do
			local num = tonumber(frame:GetName():match("PT3Bar(%d+)"))
			if num == 1 then
				xPrev = frame:GetLeft()
				yPrev = frame:GetBottom()
			end
		end
		
		for _,frame in pairs(buttons) do
		
			local frameNumber = tonumber(frame:GetName():match("PT3Bar(%d+)"))
			local count = 0
			
			if not emptyButtons[frameNumber] then
				for v in pairs( emptyButtons ) do
					if tonumber( v ) < frameNumber then
						count = count + 1
					end
				end
			end
			
			frame:ClearAllPoints()

			if PT3Bar.db.profile.compressH then
				frame:SetPoint("BOTTOMLEFT", UIParent, "BOTTOMLEFT", xPrev+((size + 4)*(frameNumber - count - 1)),yPrev)
			else
				frame:SetPoint("BOTTOMLEFT", UIParent, "BOTTOMLEFT", xPrev,yPrev-((size+4)*(frameNumber - count - 1)))	
			end		
		end
	end
end

function PT3Bar:SetTooltip(frame)
	if frame.item and self.db.profile.tooltip then
		GameTooltip_SetDefaultAnchor(GameTooltip, frame)
		GameTooltip:SetHyperlink("item:"..frame.item..":0:0:0:0:0:0:0")
	end
end

function PT3Bar:SetButtonHidden(button, v)
	local moveoverlay = button.moveoverlay
	if v then
		if not moveoverlay.hiddentex then
			moveoverlay.hiddentex = moveoverlay:CreateTexture(nil, 'BACKGROUND')
			moveoverlay.hiddentex:SetAllPoints(moveoverlay)
		end
		moveoverlay.hiddentex:SetTexture(0, 1, 0, 0.6)
	else
		if moveoverlay.hiddentex then
			moveoverlay.hiddentex:SetTexture(0, 0, 0, 0)
		end
	end
end

function PT3Bar:UpdateButton(frame,itemid,count)
	frame:SetAttribute("item", 'item:'..itemid)
	if frame.hidden then
		if locked then
			frame:Hide()
		else
			frame:Show()
		end
	else
		frame:Show()
	end
	local _,_,_,_,_,_,_,stacksize,_,texture = GetItemInfo(itemid)
	frame:SetAlpha(1)
	frame.item = itemid
	frame.icon:SetTexture(texture)
	if (not stacksize or stacksize == 1) and (not count or count <= 1) then
		frame.count:SetText('')
	else
		frame.count:SetText(count)
	end
end

function PT3Bar:UpdateCooldowns()
	for _,frame in pairs(buttons) do
		self:UpdateCooldown(frame)
	end
	for _, t in ipairs(popupbuttons) do
		for _, frame in ipairs(t) do
			self:UpdateCooldown(frame)
		end
	end
end

function PT3Bar:UpdateCooldown(frame)
	if frame.item and not frame.hidden then
		CooldownFrame_SetTimer(frame.cooldown,GetItemCooldown(frame.item))
	end
end
